/*
 * Copyright (c) 2007 IBM Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 */
package slice;

import java.util.Hashtable;

public class Slice7 {
  public static void main(String args[]) {
    @SuppressWarnings("unused")
    Hashtable<Object, Object> abc = new Hashtable<Object, Object>(3, 0.75f);
  }
}
