/**
 * This package contains an exception analysis. For an interprocedural
 * exception analysis use {@link com.ibm.wala.analysis.exceptionanalysis.ExceptionAnalysis}. 
 * If you need a CFG without unnecessary exception edges use {@link com.ibm.wala.ipa.cfg.PrunedCFG}
 * in combination with {@link com.ibm.wala.analysis.exceptionanalysis.ExceptionAnalysis2EdgeFilter} 

 */
package com.ibm.wala.analysis.exceptionanalysis;
