/*
 * Copyright (c) 2007 IBM Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 */
package slice;

import java.util.Vector;

public class Slice6 {
  public static Vector<Integer> messages;

  public static void main(String[] args) {

    messages = new Vector<Integer>();
    messages.add(5);

    int message = messages.elementAt(0);

    sendMessage(message);
  }

  public static void sendMessage(int message) {
    // do nothing
  }
}
