package com.ibm.wala.analysis.arraybounds.hypergraph;

/**
 * This package contains a generic implementation of directed hypergraphs.
 * The implementation is optimized for shortest path search, with
 * {@link com.ibm.wala.analysis.arraybounds.hypergraph.algorithms.ShortestPath#compute(com.ibm.wala.analysis.arraybounds.hypergraph.DirectedHyperGraph,
 * com.ibm.wala.analysis.arraybounds.hypergraph.HyperNode, java.util.Comparator) ShortestPath}
 */
