/*
 * Copyright (c) 2002 - 2014 IBM Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 */

package com.ibm.wala.cast.ipa.callgraph;

import com.ibm.wala.ipa.modref.ExtendedHeapModel;

public interface AstHeapModel extends ExtendedHeapModel, AstPointerKeyFactory {

}
