// I know it sounds weird, but JS in the wild contains cases in which there are about this amount of statements in the "main" scope (function).
// So, we cannot have recursions with depth linear in the lenght of the method.
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
Console.log("1")
Console.log("2")
Console.log("3")
Console.log("4")
Console.log("5")
Console.log("6")
Console.log("7")
Console.log("8")
Console.log("9")
Console.log("0")
