/*
 * Copyright (c) 2006 IBM Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 */

package com.ibm.wala.ipa.cha;

/**
 * Exception class that indicates that construction of class hierarchy has been cancelled by a progress monitor.
 */
public final class CancelCHAConstructionException extends ClassHierarchyException {

  private static final long serialVersionUID = -1987107302523285889L;

  public CancelCHAConstructionException() {
    super("class hierarchy construction was cancelled");
  }

}
