(function nodejsModule() {
	var module = {
		require: function __WALA__require(id) {},
		exports: {}
	};
	
	(function moduleSource(exports, require, module, __filename, __dirname) {
		/*/ WALA-INSERT-CODE-HERE /*/
	})(module.exports, module.require, module, '/*/ WALA-INSERT-FILENAME-HERE /*/', '/*/ WALA-INSERT-DIRNAME-HERE /*/');
	
	return module.exports;
})();