/*
 * Copyright (c) 2006 IBM Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 */
package com.ibm.wala.core.tests.util;

public interface TestConstants {

  public final static String WALA_TESTDATA = "wala.testdata.txt";

  public static final String CLASSCONSTANT_MAIN = "LclassConstant/ClassConstant";

  public static final String PI_TEST_MAIN = "Lpi/PiNodeCallGraphTestCase";

  public static final String RECURSE_MAIN = "Lrecurse/NList";

  public final static String HELLO = "hello.txt";

  public final static String HELLO_MAIN = "Lhello/Hello";

  public final static String BCEL = "bcel.txt";

  public final static String BCEL_VERIFIER_MAIN = "Lorg/apache/bcel/verifier/Verifier";

  public final static String JLEX = "JLex.txt";

  public final static String JLEX_MAIN = "LJLex/Main";

  public final static String JAVA_CUP = "java_cup.txt";

  public final static String JAVA_CUP_MAIN = "Ljava_cup/Main";

  public final static String MULTI_DIM_MAIN = "LmultiDim/TestMultiDim";

  public final static String ARRAY_ALIAS_MAIN = "LarrayAlias/TestArrayAlias";

  public final static String ZERO_LENGTH_ARRAY_MAIN = "LarrayAlias/TestZeroLengthArray";

  public final static String REFLECT1_MAIN = "Lreflection/Reflect1";

  public final static String REFLECT2_MAIN = "Lreflection/Reflect2";

  public final static String REFLECT3_MAIN = "Lreflection/Reflect3";

  public final static String REFLECT4_MAIN = "Lreflection/Reflect4";

  public final static String REFLECT5_MAIN = "Lreflection/Reflect5";

  public final static String REFLECT6_MAIN = "Lreflection/Reflect6";

  public final static String REFLECT7_MAIN = "Lreflection/Reflect7";

  public final static String REFLECT8_MAIN = "Lreflection/Reflect8";

  public final static String REFLECT9_MAIN = "Lreflection/Reflect9";

  public final static String REFLECT10_MAIN = "Lreflection/Reflect10";

  public final static String REFLECT11_MAIN = "Lreflection/Reflect11";

  public final static String REFLECT12_MAIN = "Lreflection/Reflect12";

  public final static String REFLECT13_MAIN = "Lreflection/Reflect13";

  public final static String REFLECT14_MAIN = "Lreflection/Reflect14";

  public final static String REFLECT15_MAIN = "Lreflection/Reflect15";

  public final static String REFLECT16_MAIN = "Lreflection/Reflect16";

  public final static String REFLECT17_MAIN = "Lreflection/Reflect17";

  public final static String REFLECT18_MAIN = "Lreflection/Reflect18";
  
  public final static String REFLECT19_MAIN = "Lreflection/Reflect19";
  
  public final static String REFLECT20_MAIN = "Lreflection/Reflect20";
  
  public final static String REFLECT21_MAIN = "Lreflection/Reflect21";
  
  public final static String REFLECT22_MAIN = "Lreflection/Reflect22";
  
  public final static String REFLECT23_MAIN = "Lreflection/Reflect23";

  public final static String REFLECTGETMETHODCONTEXT_MAIN = "Lreflection/GetMethodContext";
  
  public final static String SLICE1_MAIN = "Lslice/Slice1";

  public final static String SLICE2_MAIN = "Lslice/Slice2";

  public final static String SLICE3_MAIN = "Lslice/Slice3";

  public final static String SLICE4_MAIN = "Lslice/Slice4";

  public final static String SLICE5_MAIN = "Lslice/Slice5";

  public final static String SLICE7_MAIN = "Lslice/Slice7";

  public final static String SLICE8_MAIN = "Lslice/Slice8";

  public final static String SLICE9_MAIN = "Lslice/Slice9";

  public final static String SLICE_TESTCD1 = "Lslice/TestCD1";

  public final static String SLICE_TESTCD2 = "Lslice/TestCD2";

  public final static String SLICE_TESTCD3 = "Lslice/TestCD3";

  public final static String SLICE_TESTCD4 = "Lslice/TestCD4";

  public final static String SLICE_TESTCD5 = "Lslice/TestCD5";

  public final static String SLICE_TESTCD6 = "Lslice/TestCD6";

  public final static String SLICE_TESTID = "Lslice/TestId";

  public final static String SLICE_TESTARRAYS = "Lslice/TestArrays";

  public final static String SLICE_TESTFIELDS = "Lslice/TestFields";

  public final static String SLICE_TESTGLOBAL = "Lslice/TestGlobal";

  public final static String SLICE_TESTMULTITARGET = "Lslice/TestMultiTarget";

  public final static String SLICE_TESTRECURSION = "Lslice/TestRecursion";

  public final static String SLICE_TEST_PRIM_GETTER_SETTER = "Lslice/TestPrimGetterSetter";

  public final static String SLICE_TEST_PRIM_GETTER_SETTER2 = "Lslice/TestPrimGetterSetter2";

  public final static String SLICE_TESTTHIN1 = "Lslice/TestThin1";

  public final static String SLICE_TESTTHROWCATCH = "Lslice/TestThrowCatch";

  public final static String SLICE_TESTMESSAGEFORMAT = "Lslice/TestMessageFormat";

  public final static String SLICE_TESTINETADDR = "Lslice/TestInetAddr";

  public final static String SLICE_JUSTTHROW = "Lslice/JustThrow";

}
